import rclpy
from rclpy.node import Node
from sensor_msgs.msg import Image  # Pour publier les images de la caméra
from sensor_msgs.msg import CameraInfo
from nav_msgs.msg import Odometry
from sensor_msgs.msg import Range  # Utilisé pour les capteurs de distance individuels
from geometry_msgs.msg import Twist, Quaternion
from std_msgs.msg import Float32MultiArray  # Utilisé pour envoyer un tableau de distances
from cv_bridge import CvBridge  # Pour convertir les images en format ROS
import numpy as np
from controller import Robot, InertialUnit, Compass, PositionSensor
import signal
import math

class EPuck2Controller(Node):
    def __init__(self, webots_robot):
        super().__init__('epuck2_controller')
        
        # Webots robot instance
        self.robot = webots_robot
        self.timestep = int(self.robot.getBasicTimeStep())
        #self.subscriber = self.create_subscription(Range, '/distance_sensors', self.distance_callback, 10)        
        self.get_logger().info("e-puck2 controller initialized")
        
               # Récupérer les moteurs de l’e-puck
        self.left_motor = self.robot.getDevice("left wheel motor")
        self.right_motor = self.robot.getDevice("right wheel motor")
        self.left_position_sensor = self.robot.getDevice("left wheel sensor")
        self.right_position_sensor = self.robot.getDevice("right wheel sensor")
        
                
        # Activer les capteurs de position pour l'odométrie
        self.left_position_sensor.enable(self.timestep)
        self.right_position_sensor.enable(self.timestep)
        
              # Définir le mode de vitesse pour les moteurs
        self.left_motor.setPosition(float('inf'))
        self.right_motor.setPosition(float('inf'))
        
                # Initialiser les vitesses
        self.left_motor.setVelocity(0.0)
        self.right_motor.setVelocity(0.0)
        
            # Abonnement aux commandes de mouvement
        self.cmd_vel_subscriber = self.create_subscription(Twist, '/cmd_vel', self.cmd_vel_callback, 10)
        
                # Configuration de la caméra
        self.camera = self.robot.getDevice("camera")  # Nom du périphérique caméra dans Webots
        self.camera.enable(self.timestep)
        
            # Création du publisher pour l'image
        self.image_publisher = self.create_publisher(Image, '/camera/image_raw', 10)
        self.bridge = CvBridge()  # Pour convertir l'image en message ROS
        
             # Configuration des capteurs de distance
        self.distance_sensors = []
        
        sensor_names = ["ps0", "ps1", "ps2", "ps3", "ps4", "ps5", "ps6", "ps7"]  # Noms des capteurs dans Webots
       
        for name in sensor_names:
            sensor = self.robot.getDevice(name)
            sensor.enable(self.timestep)
            self.distance_sensors.append(sensor)
            
        self.distance_publisher = self.create_publisher(Float32MultiArray, '/distance_sensors', 10)


        
        
        # Publisher pour l'odométrie        
        self.odom_publisher = self.create_publisher(Odometry, '/odom', 10)
        self.x = 0.0
        self.y = 0.0
        self.theta = 0.0
      
        
        print ("E-puck ros2 started");
   
        
    def publish_distance_sensors(self):
        distances = Float32MultiArray()
        distances.data = [sensor.getValue() * 0.001 for sensor in self.distance_sensors]  # Conversion en mètres
        self.distance_publisher.publish(distances)

        
    def publish_camera_image(self):
        # Capture d'une image et conversion en format ROS
        width = self.camera.getWidth()
        height = self.camera.getHeight()
        image_data = self.camera.getImageArray()
        if image_data:
            # Convertir l'image Webots (liste 2D) en tableau NumPy
            np_image = np.array(image_data, dtype=np.uint8).reshape((height, width , 3))
            # Conversion en message Image ROS2
            image_msg = self.bridge.cv2_to_imgmsg(np_image, encoding="rgb8")
            self.image_publisher.publish(image_msg)
            
    def publish_compass_data(self):
        compass_msg = MagneticField()
        compass_msg.header.stamp = self.get_clock().now().to_msg()
        
        # Récupérer les valeurs du compas
        compass_values = self.compass.getValues()
        compass_msg.magnetic_field.x = compass_values[0]
        compass_msg.magnetic_field.y = compass_values[1]
        compass_msg.magnetic_field.z = compass_values[2]
        
        self.compass_publisher.publish(compass_msg)
    
    def publish_odometry(self):
        left_position = self.left_position_sensor.getValue()
        right_position = self.right_position_sensor.getValue()
        
        # Calcul de la position et de l'orientation
        distance_l = left_position * 0.02  # Ajustez en fonction du rayon de roue
        distance_r = right_position * 0.02
        delta_d = (distance_r + distance_l) / 2.0
        delta_theta = (distance_r - distance_l) / 0.053  # Distance entre les roues
        
        self.x += delta_d * math.cos(self.theta)
        self.y += delta_d * math.sin(self.theta)
        self.theta += delta_theta
        
        odom_msg = Odometry()
        odom_msg.header.stamp = self.get_clock().now().to_msg()
        odom_msg.pose.pose.position.x = self.x
        odom_msg.pose.pose.position.y = self.y
        odom_msg.pose.pose.orientation = Quaternion(
            x=0.0,
            y=0.0,
            z=math.sin(self.theta / 2.0),
            w=math.cos(self.theta / 2.0)
        )
        
        self.odom_publisher.publish(odom_msg)
        
        
       # self.timer = self.create_timer(0.1, self.timer_callback)  # Call every 0.1 seconds


    def cmd_vel_callback(self, msg):
        # Conversion des commandes Twist en vitesses de roues
        # Pour un différentiel simple (comme e-puck), vitesse_gauche et vitesse_droite peuvent être calculées comme suit:
        vitesse_gauche = msg.linear.x - msg.angular.z
        vitesse_droite = msg.linear.x + msg.angular.z

        # Appliquer les vitesses aux moteurs
        self.left_motor.setVelocity(vitesse_gauche)
        self.right_motor.setVelocity(vitesse_droite)
        

    def publish_compass_data(self):
        compass_msg = MagneticField()
        compass_msg.header.stamp = self.get_clock().now().to_msg()
        
        # Récupérer les valeurs du compas
        compass_values = self.compass.getValues()
        compass_msg.magnetic_field.x = compass_values[0]
        compass_msg.magnetic_field.y = compass_values[1]
        compass_msg.magnetic_field.z = compass_values[2]
        
        self.compass_publisher.publish(compass_msg)
    
        
        
    def timer_callback(self):
        print("Send", 0.1)  # Example print statement or publishing to a topic
        self.get_logger().info("Timer")

def signal_handler(signum, frame):
    rclpy.shutdown()


def main(args=None):
    signal.signal(signal.SIGINT, signal_handler)
    robot = Robot()  # Initialisez l'instance du robot Webots
    rclpy.init(args=args)
    controller = EPuck2Controller(robot)
    print("Robot step: ", controller.timestep)

    while (robot.step( controller.timestep))!=-1:
        rclpy.spin_once(controller, timeout_sec=0) 
        controller.publish_camera_image()  # Publier l'image de la caméra à chaque étape
        controller.publish_distance_sensors()  # Publier les données des capteurs de distance
        controller.publish_odometry()
                 
    controller.destroy_node()
    rclpy.shutdown()

if __name__ == '__main__':
    main()
